package com.fetchLocal;

import android.content.Context;
import android.net.ConnectivityManager;
import android.net.Network;
import android.net.NetworkCapabilities;
import android.util.Log;

import com.facebook.react.bridge.Arguments;
import com.facebook.react.bridge.Promise;
import com.facebook.react.bridge.ReactApplicationContext;
import com.facebook.react.bridge.ReadableMap;
import com.facebook.react.bridge.ReadableMapKeySetIterator;
import com.facebook.react.bridge.WritableMap;
import com.fetchLocal.functional.BiFunction;
import com.fetchLocal.handler.HttpMethodHandlerFactory;

import org.apache.commons.io.IOUtils;

import java.io.IOException;
import java.io.InputStream;
import java.io.StringWriter;
import java.net.HttpURLConnection;
import java.net.URL;

public class FetchLocalRunnable implements Runnable {

    private static final String METHOD_FIELD_NAME = "method";

    private static final String CONTENT_TYPE_HEADER = "Content-Type";

    private static final String HEADERS_FIELD_NAME = "headers";

    private static final String RESPONSE_KEY = "response";

    private static final String TIMEOUT = "timeout";

    private static final int DEFAULT_TIMEOUT = 5000;

    private ReactApplicationContext context;

    private String link;

    private ReadableMap data;

    private Promise promise;

    public FetchLocalRunnable(ReactApplicationContext context, String link, ReadableMap data, Promise promise) {
        this.context = context;
        this.link = link;
        this.data = data;
        this.promise = promise;
    }

    @Override
    public void run() {
        try {
            Log.i(getClass().getName(), String.format("Fetching On Network: %s", link));
            Network network = getWifiNetwork(context);
            ReadableMap result = fetchImpl(network, link, data);
            promise.resolve(result);
        } catch (Exception e) {
            Log.e(getClass().getName(), String.format("Error Fetching link: %s", link), e);
            promise.reject(e);
        }
    }

    protected Network getWifiNetwork(ReactApplicationContext context) {
        final ConnectivityManager cm = (ConnectivityManager) context.getSystemService(Context.CONNECTIVITY_SERVICE);
        Network networkList[] = cm.getAllNetworks();
        for (Network network : networkList) {
            NetworkCapabilities info = cm.getNetworkCapabilities(network);
            if (info.hasTransport(NetworkCapabilities.TRANSPORT_WIFI)) {
                return network;
            }
        }
        throw new RuntimeException("Unable to find any network matching the network criteria for fetch local");
    }

    protected ReadableMap fetchImpl(Network network, String link, ReadableMap data) throws Exception {
        HttpURLConnection connection = null;
        try {
            Log.i(FetchLocalModule.class.getName(), String.format("Url: %s", link));
            URL url = new URL(link);
            connection = (HttpURLConnection) network.openConnection(url);
            connection.setDoInput(true);
            int timeout = getTimeout(data, DEFAULT_TIMEOUT);
            connection.setReadTimeout(timeout);
            connection.setConnectTimeout(timeout);
            ReadableMap headers = data.getMap(HEADERS_FIELD_NAME);
            setHeaders(connection, headers);
            String method = data.getString(METHOD_FIELD_NAME);
            Log.d(getClass().getName(), method);
            connection.setRequestMethod(method);
            BiFunction<HttpURLConnection, ReadableMap, Void> methodHandler =
                    HttpMethodHandlerFactory.getMethodHandler(method, headers.getString(CONTENT_TYPE_HEADER));
            methodHandler.apply(connection, data);
            String response = connectAndReceiveResponse(connection);
            Log.i(getClass().getName(), String.format("Success: %s", response));
            WritableMap map = Arguments.createMap();
            map.putString(RESPONSE_KEY, response);
            return map;
        } finally {
            if (connection != null)
                connection.disconnect();
        }
    }

    protected int getTimeout(ReadableMap data, int defaultVal) {
        int timeout = data.getInt(TIMEOUT);
        if (timeout <= 0) timeout = defaultVal;
        return timeout;
    }

    protected void setHeaders(HttpURLConnection connection, ReadableMap headers) {
        ReadableMapKeySetIterator iterator = headers.keySetIterator();
        while (iterator.hasNextKey()) {
            String key = iterator.nextKey();
            connection.setRequestProperty(key, headers.getString(key));
        }
        Log.d(getClass().getName(), headers.getString(CONTENT_TYPE_HEADER));
    }

    protected String connectAndReceiveResponse(HttpURLConnection connection) throws
            IOException {
        InputStream stream = null;
        try {
            connection.connect();
            stream = connection.getInputStream();
            StringWriter writer = new StringWriter();
            IOUtils.copy(stream, writer, "UTF-8");
            String response = writer.toString();
            Log.d(getClass().getName(), "Response" + response);
            return response;
        } finally {
            IOUtils.closeQuietly(stream);
        }
    }

}
