package com.fetchLocal;

import android.content.Context;
import android.net.ConnectivityManager;
import android.net.Network;
import android.net.NetworkCapabilities;

import com.facebook.react.bridge.Arguments;
import com.facebook.react.bridge.JavaOnlyMap;
import com.facebook.react.bridge.Promise;
import com.facebook.react.bridge.ReactApplicationContext;
import com.facebook.react.bridge.ReadableMap;
import com.fetchLocal.handler.HttpMethodHandlerFactory;
import com.fetchLocal.handler.PutPostBodyMethodHandler;
import com.google.common.collect.Lists;

import org.apache.commons.collections4.MapUtils;
import org.apache.commons.io.IOUtils;
import org.apache.commons.lang3.StringUtils;
import org.junit.After;
import org.junit.Assert;
import org.junit.Before;
import org.junit.Test;
import org.junit.runner.RunWith;
import org.mockito.Mockito;
import org.mockito.MockitoAnnotations;
import org.mockito.internal.util.StringUtil;
import org.mockito.invocation.InvocationOnMock;
import org.mockito.stubbing.Answer;
import org.powermock.api.mockito.PowerMockito;
import org.powermock.core.classloader.annotations.PrepareForTest;
import org.powermock.modules.junit4.PowerMockRunner;

import java.io.ByteArrayInputStream;
import java.io.IOException;
import java.io.InputStream;
import java.net.HttpURLConnection;
import java.net.URL;
import java.util.List;

import static org.junit.Assert.*;
import static org.mockito.ArgumentMatchers.any;
import static org.mockito.ArgumentMatchers.anyString;
import static org.mockito.ArgumentMatchers.eq;
import static org.mockito.Mockito.atLeastOnce;
import static org.mockito.Mockito.doNothing;
import static org.mockito.Mockito.doReturn;
import static org.mockito.Mockito.doThrow;
import static org.mockito.Mockito.mock;
import static org.mockito.Mockito.spy;
import static org.mockito.Mockito.times;
import static org.mockito.Mockito.verify;
import static org.mockito.Mockito.when;

@RunWith(PowerMockRunner.class)
@PrepareForTest({Network[].class, HttpMethodHandlerFactory.class, Arguments.class})
public class FetchLocalRunnableTest {

    private static final String link = "http://192.168.0.104:3000/home";

    private static final int DEFAULT_VAL = 5000;

    private static ReactApplicationContext context;

    private ReadableMap data;

    private static ReadableMap getHeadersData() {
        List<Object> objectListForHeaders = Lists.newArrayList();
        objectListForHeaders.add("Accept");
        objectListForHeaders.add("application/json");
        objectListForHeaders.add("Content-Type");
        objectListForHeaders.add("application/json");
        return JavaOnlyMap.of(objectListForHeaders.toArray());
    }

    private static ReadableMap getData(String body, int timeout, String method, ReadableMap headersData) {
        List<Object> objectList = Lists.newArrayList();
        objectList.add("body");

        objectList.add(StringUtils.isBlank(body) ? "{\"id\":\"appo_id\"}" : body);
        objectList.add("timeout");
        objectList.add(timeout == Integer.MIN_VALUE ? 4000 : timeout);
        objectList.add("method");
        objectList.add(StringUtils.isBlank(method) ? "POST" : method);
        objectList.add("headers");
        objectList.add(MapUtils.isEmpty(headersData.toHashMap()) ? getHeadersData() : headersData);
        return JavaOnlyMap.of(objectList.toArray());
    }

    private static ReadableMap getData() {
        return getData(null, Integer.MIN_VALUE, null, new JavaOnlyMap());
    }


    @Before
    public void setUp() throws Exception {
        MockitoAnnotations.initMocks(this);
        context = mock(ReactApplicationContext.class);
        data = getData();
    }

    @After
    public void tearDown() throws Exception {
    }

    @Test
    public void testRunSuccess() throws Exception {
        Promise promise = mock(Promise.class);
        Network network = Mockito.mock(Network.class);

        FetchLocalRunnable fetchLocalRunnableSpy = spy(new FetchLocalRunnable(context, link, data, promise));
        doReturn(network).when(fetchLocalRunnableSpy).getWifiNetwork(context);
        doReturn(data).when(fetchLocalRunnableSpy).fetchImpl(network, link, data);

        fetchLocalRunnableSpy.run();

        verify(fetchLocalRunnableSpy, times(1)).getWifiNetwork(context);
        verify(fetchLocalRunnableSpy, times(1)).fetchImpl(network, link, data);
        verify(promise).resolve(data);
    }

    @Test
    public void testRunException() throws Exception {
        Promise promise = mock(Promise.class);
        Network network = Mockito.mock(Network.class);

        FetchLocalRunnable fetchLocalRunnableSpy = spy(new FetchLocalRunnable(context, link, data, promise));
        doReturn(network).when(fetchLocalRunnableSpy).getWifiNetwork(context);

        Exception e = new RuntimeException();
        doThrow(e).when(fetchLocalRunnableSpy).fetchImpl(network, link, data);

        fetchLocalRunnableSpy.run();

        verify(fetchLocalRunnableSpy, times(1)).getWifiNetwork(context);
        verify(fetchLocalRunnableSpy, times(1)).fetchImpl(network, link, data);
        verify(promise).reject(e);
    }

    @Test
    public void testGetWifiNetworkSuccess() {
        Promise promise = mock(Promise.class);
        Network network = Mockito.mock(Network.class);
        Network networkList[] = {network};
        NetworkCapabilities networkCapabilities = mock(NetworkCapabilities.class);
        ConnectivityManager connectivityManager = mock(ConnectivityManager.class);

        when(context.getSystemService(Context.CONNECTIVITY_SERVICE)).thenReturn(connectivityManager);
        when(connectivityManager.getAllNetworks()).thenReturn(networkList);
        when(connectivityManager.getNetworkCapabilities(network)).thenReturn(networkCapabilities);
        when(networkCapabilities.hasTransport(NetworkCapabilities.TRANSPORT_WIFI)).thenReturn(true);

        FetchLocalRunnable fetchLocalRunnableSpy = spy(new FetchLocalRunnable(context, link, data, promise));
        Network actualNetwork = fetchLocalRunnableSpy.getWifiNetwork(context);

        assertEquals(network, actualNetwork);
        verify(networkCapabilities, times(1)).hasTransport(NetworkCapabilities.TRANSPORT_WIFI);
    }

    @Test
    public void testGetWifiNetworkError() {
        Promise promise = mock(Promise.class);
        Network network = Mockito.mock(Network.class);
        Network networkList[] = {network};
        NetworkCapabilities networkCapabilities = mock(NetworkCapabilities.class);
        ConnectivityManager connectivityManager = mock(ConnectivityManager.class);

        when(context.getSystemService(Context.CONNECTIVITY_SERVICE)).thenReturn(connectivityManager);
        when(connectivityManager.getAllNetworks()).thenReturn(networkList);
        when(connectivityManager.getNetworkCapabilities(network)).thenReturn(networkCapabilities);
        when(networkCapabilities.hasTransport(NetworkCapabilities.TRANSPORT_WIFI)).thenReturn(false);

        FetchLocalRunnable fetchLocalRunnableSpy = spy(new FetchLocalRunnable(context, link, data, promise));

        Throwable exception = assertThrows(RuntimeException.class, () -> fetchLocalRunnableSpy.getWifiNetwork(context));
        verify(networkCapabilities, times(1)).hasTransport(NetworkCapabilities.TRANSPORT_WIFI);
        assertEquals("Unable to find any network matching the network criteria for fetch local", exception.getMessage());
    }

    @Test
    public void fetchImplTest() throws Exception {
        HttpURLConnection connection = Mockito.mock(HttpURLConnection.class);
        Promise promise = Mockito.mock(Promise.class);
        Network network = mock(Network.class);
        int timeout = data.getInt("timeout");
        String methodType = data.getString("method");
        URL url = new URL(link);
        String message = "Welcome to " + data.getString("method");
        FetchLocalRunnable fetchLocalRunnableSpy = spy(new FetchLocalRunnable(context, link, data, promise));

        when(network.openConnection(url)).thenReturn(connection);
        PutPostBodyMethodHandler putPostBodyMethodHandler = mock(PutPostBodyMethodHandler.class);
        PowerMockito.mockStatic(HttpMethodHandlerFactory.class);
        PowerMockito.when(HttpMethodHandlerFactory.getMethodHandler(eq("POST"), Mockito.anyString())).thenReturn(putPostBodyMethodHandler);
        PowerMockito.mockStatic(Arguments.class);
        PowerMockito.when(Arguments.createMap()).thenAnswer(new Answer<Object>() {
            @Override
            public Object answer(InvocationOnMock invocation) throws Throwable {
                return new JavaOnlyMap();
            }
        });
        doReturn(message).when(fetchLocalRunnableSpy).connectAndReceiveResponse(connection);


        ReadableMap result = fetchLocalRunnableSpy.fetchImpl(network, link, data);

        verify(network).openConnection(url);
        Assert.assertTrue(message.equals(result.getString("response")));
        verify(connection).setDoInput(true);
        verify(fetchLocalRunnableSpy).getTimeout(data, DEFAULT_VAL);
        verify(connection).setReadTimeout(timeout);
        verify(connection).setConnectTimeout(timeout);
        verify(fetchLocalRunnableSpy).setHeaders(connection, getHeadersData());
        verify(connection).setRequestMethod(methodType);
        PowerMockito.verifyStatic(HttpMethodHandlerFactory.class, Mockito.times(1));
        HttpMethodHandlerFactory.getMethodHandler(eq("POST"), Mockito.anyString());
        verify(putPostBodyMethodHandler).apply(connection, data);
        verify(fetchLocalRunnableSpy).connectAndReceiveResponse(connection);
        PowerMockito.verifyStatic(Arguments.class, Mockito.times(1));
        Arguments.createMap();
    }

    @Test
    public void testFetchImplException() throws Exception {
        HttpURLConnection connection = Mockito.mock(HttpURLConnection.class);
        Promise promise = Mockito.mock(Promise.class);
        Network network = mock(Network.class);
        int timeout = data.getInt("timeout");
        String methodType = data.getString("method");
        URL url = new URL(link);
        String message = "Welcome to " + data.getString("method");
        FetchLocalRunnable fetchLocalRunnableSpy = spy(new FetchLocalRunnable(context, link, data, promise));

        when(network.openConnection(url)).thenReturn(connection);
        PutPostBodyMethodHandler putPostBodyMethodHandler = mock(PutPostBodyMethodHandler.class);
        PowerMockito.mockStatic(HttpMethodHandlerFactory.class);
        PowerMockito.when(HttpMethodHandlerFactory.getMethodHandler(eq("POST"), Mockito.anyString()))
                .thenThrow(new RuntimeException("Error"));

        Throwable exception = assertThrows(RuntimeException.class, () -> fetchLocalRunnableSpy.fetchImpl(network, link, data));
        assertEquals("Error", exception.getMessage());
        verify(network).openConnection(url);
        verify(connection).disconnect();
    }

    @Test
    public void getTimeoutTest() {
        Promise promise = Mockito.mock(Promise.class);

        FetchLocalRunnable fetchLocalRunnableSpy = spy(new FetchLocalRunnable(context, link, data, promise));

        ReadableMap data = getData();
        assertEquals(data.getInt("timeout"), fetchLocalRunnableSpy.getTimeout(data, DEFAULT_VAL));

        data = getData(null, 0, null, new JavaOnlyMap());
        assertEquals(DEFAULT_VAL, fetchLocalRunnableSpy.getTimeout(data, DEFAULT_VAL));

        data = getData(null, -10, null, new JavaOnlyMap());
        assertEquals(DEFAULT_VAL, fetchLocalRunnableSpy.getTimeout(data, DEFAULT_VAL));
    }

    @Test
    public void setHeadersTest() {
        Promise promise = Mockito.mock(Promise.class);
        HttpURLConnection connection = Mockito.mock(HttpURLConnection.class);
        FetchLocalRunnable fetchLocalRunnableSpy = spy(new FetchLocalRunnable(context, link, data, promise));
        fetchLocalRunnableSpy.setHeaders(connection, getHeadersData());
        verify(connection, atLeastOnce()).setRequestProperty(anyString(), anyString());
    }

    @Test
    public void connectAndReceiveResponseTest() throws IOException {
        Promise promise = Mockito.mock(Promise.class);
        FetchLocalRunnable fetchLocalRunnableSpy = spy(new FetchLocalRunnable(context, link, data, promise));
        HttpURLConnection connection = Mockito.mock(HttpURLConnection.class);
        String message = "Welcome to " + data.getString("method");
        InputStream inputStream = new ByteArrayInputStream(message.getBytes());
        when(connection.getInputStream()).thenReturn(inputStream);
        String response = fetchLocalRunnableSpy.connectAndReceiveResponse(connection);
        assertTrue(message.equals(response));
        verify(connection).connect();
        verify(connection).getInputStream();
    }

}